# coding: utf-8
"""

    ファイル名：   DummyBridge.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""

import os
import re
import numpy                as np
from scipy                  import stats, optimize
from bisect                 import bisect_right
import matplotlib.pyplot    as plt
# import seaborn              as sns
from lmfit                  import minimize, Parameters, Parameter, report_fit
from IntensityData          import IntensityData
from SimulatedAnnealing     import SimulatedAnnealing, animate_annealing

DEBUG = False

class DummyBridge:
    def __init__( self, Q, I, smoother=None ):
        self.Q  = Q
        self.I  = I

    def get_point_index( self, q ):
        return max( 0, min( self.Q.shape[0]-1, bisect_right( self.Q, q ) ) )

    def fit( self, q1, q2 ):

        self.q1 = q1
        self.q2 = q2
        self.p1 = p1 = self.get_point_index( q1 )
        self.p2 = p2 = self.get_point_index( q2 )
        self.params = []

        for i1, i2 in [ [0, p1], [ p1, p2 ], [ p2, -1 ] ]:
            x1 = self.Q[i1]
            x2 = self.Q[i2]
            if i1 == 0:
                y1 = np.max( self.I )
            else:
                y1 = self.I[i1]
            y2 = self.I[i2]
            slope       = ( y2 - y1 ) / ( x2 - x1 )
            intercept   = y2 - slope * x2
            self.params.append( [ slope, intercept ] )

    def predict( self, x ):
        q1_q2 = np.logical_and( x >= self.q1, x < self.q2  )
        y_  = np.hstack( [  self.params[0][0] * x[ x < self.q1 ]    + self.params[0][1],    # Guinier
                            self.params[1][0] * x[ q1_q2]           + self.params[1][1],    # Bridge
                            self.params[2][0] * x[ x >= self.q2 ]   + self.params[2][1],    # Porod
                         ] )
        return y_

def demo_plot( file, q1, q2 ):

    intensity = IntensityData( file )
    d, f = os.path.split( file )

    plt.title( 'Bridged Guinier-Porod demo for ' + f )

    plt.plot( intensity.Q, intensity.I )
    plt.xlabel( 'Q' )
    plt.ylabel( 'Intensity' )

    bridge = DummyBridge( intensity.Q, intensity.I )
    bridge.fit( q1, q2 )

    for p in [bridge.p1, bridge.p2]:
        plt.plot( intensity.Q[p], intensity.I[p], 'o', color='red'  )

    x = intensity.Q[bridge.p1:bridge.p2]
    y = bridge.predict( x )
    plt.plot( x, y, color='pink' )

    plt.show()

def demo_anneal( file, sq1, sq2 ):

    intensity   = IntensityData( file )
    dummy       = DummyBridge( intensity.Q, intensity.I )

    def f( x ):
        dummy.fit( x[0], x[1] )
        y_ = dummy.predict( intensity.Q )
        return np.log( np.sum( ( intensity.W * (intensity.I - y_) )**2 ) )

    def xconstaints( x ):
        return x[0] < x[1]

    anneal = SimulatedAnnealing()
    anneal.minimize( f, xrange=[ [0.0, 0.5], [0.0, 0.5] ], start=[sq1, sq2], seed=1234,
                        n = 200,
                        xconstaints=xconstaints )

    print( 'anneal.start=', anneal.start )

    i1 = np.linspace( 0.0, 0.5, 101 )
    i2 = np.linspace( 0.0, 0.5, 101 )

    fmax = f( [ 0.001, 0.9 ] ) * 2

    x1m, x2m = np.meshgrid(i1, i2)
    fm = np.zeros(x1m.shape)
    for i in range(x1m.shape[0]):
        for j in range(x1m.shape[1]):
            q1 = x1m[i][j]
            q2 = x2m[i][j]
            if q1 < q2:
                fm[i][j] = f( [ q1, q2 ] )
            else:
                fm[i][j] = fmax

    demo_plot( file, anneal.xc[0], anneal.xc[1] )

    animate_annealing( (x1m, x2m, fm), anneal )
