# coding: utf-8
"""
    UsableRange.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""
import time
from bisect                 import bisect_right
import numpy                as np
from lmfit                  import minimize, Parameters, Parameter, report_fit
from DataModelsAgg          import _GuinierPorodAgg
# from NumpyUtils             import np_savetxt
from GaussianProcessDep     import Smoother

DEBUG = True

ACCEPTABLE_X_DEVIATION      = 1.5

class UsableRange:
    def __init__( self, intensity, model_class=None, debugger_queues=None ):

        self.intensity = intensity
        self.x  = intensity.Q
        self.y  = intensity.I
        self.w  = intensity.W
        self.X  = intensity.X
        self.Y  = intensity.Y
        self.model_class = model_class
        self.debugger_queues = debugger_queues

    def get_range( self, proof_plot=False ):

        if proof_plot:
            import matplotlib.pyplot    as plt
            from matplotlib.ticker      import FuncFormatter
            from scipy                  import stats
            start_ = time.time()

        # if not proof_plot and not self.intensity.basically_ok:
        #     raise Exception( "Can't get range" )

        f1, t1 = self.intensity.f1, self.intensity.t1

        if DEBUG:  print( 'f1, t1=', [ f1, t1 ] )

        tx = None
        if self.model_class is not None:
            t0_ = time.time()
            if proof_plot: print( 'compute_basic_range took ', t0_ - start_ )

            self.params = None

            try:
                _, tx = self.extend( f1, t1 )
            except:
                from ExceptionTracebacker   import ExceptionTracebacker
                etb = ExceptionTracebacker()
                print(etb)
                tx = t1
            # te_ = time.time()
            # print( 'it took ', te_ - t0_ )
            # print( 'extended f1, tx=', [ f1, tx ] )

        if proof_plot:
            p_array = self.intensity.p_array
            s_array = self.intensity.slope_array
            f10     = self.intensity.f10
            t10     = self.intensity.t10
            # np_savetxt( 'temp.csv', eval_array )
            fig    = plt.figure( figsize=(24,8) )
            ax1 = fig.add_subplot( 131 )
            ax2 = fig.add_subplot( 132 )
            ax3 = fig.add_subplot( 133 )
            ax1.set_title( 'Slopes' )
            ax2.set_title( 'P Values' )
            ax3.set_title( 'Intensities' )
            ax3.xaxis.set_major_formatter( FuncFormatter( self.x_formatter ) )
            ax1.plot( s_array )
            s_array_pa  = s_array[ self.intensity.p_acceptable ]
            a_array_i   = self.intensity.p_ok
            ax1.plot( a_array_i, s_array_pa, 'r:' )
            ax2.plot( p_array )
            ymin1, ymax1 = ax2.get_ylim()
            ax2.plot( [ f10, f10 ], [ ymin1, ymax1 ], 'b:' )
            ax2.plot( [ t10, t10 ], [ ymin1, ymax1 ], 'g:' )
            ax3.scatter( self.X, self.Y )
            ymin3, ymax3 = ax3.get_ylim()
            ax3.set_ylim( ymin3, ymax3 )
            f1x = self.X[ f1 ]
            t1x = self.X[ t1 ]
            ax3.plot( [ f1x, f1x ], [ ymin3, ymax3 ], 'b:' )
            ax3.plot( [ t1x, t1x ], [ ymin3, ymax3 ], 'g:' )
            if tx is not None:
                tx_ = tx / 10
                ax2.plot( [ tx_, tx_ ], [ ymin1, ymax1 ], 'r:' )
                txx = self.X[ tx ]
                ax3.plot( [ txx, txx ], [ ymin3, ymax3 ], 'r:' )

                xmin3, xmax3 = ax3.get_xlim()
                xoffset = ( xmax3 - xmin3 ) * 0.05
                yoffset = ( ymax3 - ymin3 ) * 0.05

                for x_, i_, xo_ in [ [ t1x, t1, -xoffset*3 ], [ txx, tx, xoffset ] ]:
                    text_ = '%.2g (%d)' % ( x_, i_ )
                    color_ = 'black'
                    txy = ymin3 + yoffset
                    ax3.annotate( text_, xy=( x_, txy ),
                                    xytext=( x_ + xo_, txy + yoffset ),
                                    color=color_, alpha=0.5,
                                    arrowprops=dict( headwidth=5, width=1, color=color_, shrink=0.05 ),
                                    )
                if  self.params is not None:
                    G1, G2, Rg1, Rg2, d1, d2 = self.params
                    Y_  = np.log( _GuinierPorodAgg( [G1, G2], [Rg1, Rg2], [d1, d2], self.x ) )
                    ax3.plot( self.X, Y_, color='pink' )

            """
            quarter_ = self.X.shape[0]//4
            result = self.intensity.evaluate_range( slice( 0, quarter_ ) )
            X4 = np.array( [ 0, self.X[-1] ] )
            Y4 = result[0] * X4 + result[1]
            ax3.plot( X4, Y4, color='cyan' )
            """

            plt.tight_layout()
            plt.show()

        if proof_plot:
            end_ = time.time()
            t_ =  end_ - start_
            print( 'it took ', t_ )

        if proof_plot and not self.intensity.basically_ok:
            raise Exception( "Can't get range" )

        if tx is not None:
            t1 = tx

        return [ f1, t1 ]

    def x_formatter( self, val, pos ):
        if pos is None: return ''

        if val >= 0:
            index = bisect_right( self.X, val )
            return '%g\n(%g)' % ( val, index )
        else:
            return '%g' % val

    def extend( self, f, t ):

        self.forced_not_to_extend = False   # this is currently used for IpI test while alternatives should be sought
        self.I0_est = None
        if True:
            # params = self.extend_impl_sub( f, t//2, solve_only=True )
            params = self.extend_impl_sub( f, t, solve_only=True )
            if params is None:
                params = self.extend_impl_sub( f, t, slice_ok=False, solve_only=True )
                if params is None:
                    print( 'extend failed!' )
                    return f, t

            Rg1, d1 = params[2], params[4]
            q1 = 1/Rg1 * np.sqrt( 3*d1/2 )
            t_ = bisect_right( self.x, q1 )
            I0_init = params[0] + params[1]
            I0_est  = self.extimate_I0( 0, t_, I0_init )
            if DEBUG: print( 'I0_init=', I0_init, ', I0_est=', I0_est, ', t_=', t_ )
            self.I0_est = I0_est

        try:
            f_, t_ = self.extend_impl_main( f, t )
        except Exception as exc:
            if DEBUG: print( '(1) exc=', exc )
            if self.debugger_queues is not None:
                if str(exc).find( 'stop' ) >= 0:
                    self.debugger_queues[1].put( [ 'Stop', None, None ] )
                raise Exception( 'terminate' )

            # self.I0_est must be wrong
            if self.I0_est is None:
                raise Exception( 'self.I0_est must be wrong!' )
            else:
                self.I0_est = None
                try:
                    f_, t_ = self.extend_impl_main( f, t )
                except Exception as exc:
                    if DEBUG: print( '(2) exc=', exc )
                    raise Exception( 'extend' )

        if self.debugger_queues is not None:
            self.model.evaluate( *self.params, debug=True, caller='main' )
            dbcmd = self.debugger_queues[0].get( block=True )
            self.debugger_queues[1].put( [ 'Stop', None, None ] )

        extend_ratio = ( t_ - f_ ) / ( t - f )
        if DEBUG: print( 'extend_ratio=', extend_ratio )
        if extend_ratio > 1.5:
            if DEBUG:
                print( 'f, t=', [ f, t ],  '; f_, t_=', [ f_, t_ ] )
                print( 'force not to extend' )
            # this is currently used for IpI test while alternatives should be sought
            self.forced_not_to_extend = True
            f_, t_ = f, t

        return f_, t_

    def extend_impl_main( self, f, t ):

        opt_result  = None
        opt_ee      = None
        opt_i       = None

        # this loop is neccesary for TY/TY03s304_ave0-4_sub.dat, etc.
        for i in range(3):

            try:
                f, t, ee, params = self.extend_impl_sub( f, t )
            except Exception as exc:
                    if DEBUG: print( '(3) exc=', exc )
                    if str(exc) == 'Stop':
                        raise Exception( 'stop in the first extend_impl_sub call' )
                    try:
                        f, t, ee, params = self.extend_impl_sub( f, t, slice_ok=False )
                    except Exception as exc:
                        if DEBUG: print( '(4) exc=', exc )
                        raise Exception( 'stop in the second extend_impl_sub call' )

            # print( "extend: %d %s %.4g" % ( i, [f, t], ee ) )
            if opt_ee is None or ee < opt_ee:
                opt_result  = [ f, t, ee, params ]
                opt_ee      = ee
                opt_i       = i

        self.ee     = opt_ee
        self.params = opt_result[3]
        self.extend_opt_i = opt_i

        if self.I0_est is None:
            self.I0_est = self.params[0] + self.params[1]
            # TODO: should retry?
            # got worse for aggregation/'exampleB_1mg.dat
            # return self.extend_impl_main( f, t )

        return opt_result[0], opt_result[1]

    def extend_impl_sub( self, f, t, slice_ok=True, solve_only=False, with_smoother=True ):

        if slice_ok:
            slice_ = slice( f, t+1 )
            x_ = self.x[ slice_ ]
            y_ = self.y[ slice_ ]
            w_ = self.w[ slice_ ]
        else:
            x_ = self.x
            y_ = self.y
            w_ = self.w

        if with_smoother:
            smoother = Smoother( x_, y_ )
            y_ = smoother( x_ )

        self.model = self.model_class( x_, y_, w_, debugger_queues=self.debugger_queues, I0_est=self.I0_est )

        xmax = self.x[t]
        Q2 = xmax * 3/4
        Q1 = Q2/2

        ee, qq, params = self.model.solve_iter( Q1, Q2 )

        if solve_only:
            return params

        initial_ee = self.evaluate_tx( params, t )

        extended_t = None
        # TODO: binary control
        lower = last_t = t
        upper = self.x.shape[0] - 1
        while True:
            t_ = ( lower + upper ) // 2
            if t_ == last_t:
                break

            # print( 'extend_impl_sub: %d = ( %d + %d ) // 2' % ( t_, lower, upper ) )
            ee_ = self.evaluate_tx( params, t_ )
            # print( t_, ee_ 
            last_t = t_
            if ee_ / initial_ee > ACCEPTABLE_X_DEVIATION:
                upper = t_
            else:
                extended_t = t_
                size = upper - lower
                if size == 0:
                    break
                else:
                    lower = t_

        if extended_t is not None:
            t = extended_t

        return f, t, ee, params

    def evaluate_tx( self, params, t_ ):
        G1, G2, Rg1, Rg2, d1, d2 = params
        slice_ = slice( t_, t_+20 )
        x_  = self.x[slice_]
        Y_  = np.log( _GuinierPorodAgg( [G1, G2], [Rg1, Rg2], [d1, d2], x_ ) )
        e = self.w[slice_] * ( self.Y[slice_] - Y_ )
        ee_ = np.inner( e, e ) / ( x_.shape[0] - 3 )
        return ee_

    def I0_model( self, params, X, Y, W ):
        G   = params[ 'G'  ].value
        Rg  = params[ 'Rg' ].value
        e = W * ( Y - np.log( G * np.exp( -X * Rg**2/3 ) ) )
        return e

    def extimate_I0( self, f, t, I0_init ):

        params  = Parameters()
        params.add('G',  value= I0_init,  min=1e-5, max=100 )
        params.add('Rg', value= 30,  min=1e-5, max=100 )

        slice_ = slice(f, t+1)
        X_ = self.X[ slice_ ]
        Y_ = self.Y[ slice_ ]
        w_ = self.w[ slice_ ]

        result = minimize( self.I0_model, params, args=( X_, Y_, w_ ) )

        G = result.params[ 'G' ].value

        self.estimate_I0_result = result

        return G
