# coding: utf-8
"""

    ExcelCOM.py - a minimal Python wrapper for Excel ole object

    Copyright (c) 2016-2017, Masatsuyo Takahashi, KEK-PF

"""
import os
import re
import numpy            as np
from win32com.client    import Dispatch, DispatchEx, gencache
import pythoncom
from shutil             import copyfile

DEBUG = False

msoTrue                         = -1
msoFalse                        =  0
msoConnectorStraight            =  1
msoShapeOval                    =  9
msoShapeRightBrace              = 32
msoTextOrientationHorizontal    =  1
msoThemeColorAccent1            = 5
msoThemeColorText1              = 13
xlCategory                      = 1
xlValue                         = 2
xlLine                          = 4
xlUpward                        = -4171
xlLow                           = -4134
xlNone                          = -4142

worksheet_path_re = re.compile('^(.+\.xls[xm])\((.+)\)$')

def CoInitialize():
    pythoncom.CoInitialize()

def CoUninitialize ():
    pythoncom.CoUninitialize()

def ColAddr( col ):
    col_ = col - 1
    addr = ''
    while True:
        r = col_ % 26
        col_ //= 26
        addr = chr(65+r) + addr
        if col_ == 0:
            break
        else:
            col_ -= 1
    return addr

def CellAddr( cell, absref='$$' ):
    row_abs = absref[0].replace( ' ', '' )
    col_cabs = absref[1].replace( ' ', '' )
    return row_abs + ColAddr( cell[1] ) + col_cabs + str(cell[0])

CLSID = 'Excel.Application'
INVOKE_EXCEL_SIMPLE_DISPATCH = True

class ExcelComClient:
    def __init__( self, new_instance=True, excel_id=None ):
        self.clsid = None
        if excel_id is None:
            if INVOKE_EXCEL_SIMPLE_DISPATCH:
                if new_instance:
                    self.excel  = DispatchEx(CLSID)
                else:
                    self.excel  = Dispatch(CLSID)
            else:
                """
                    borrowed from:
                    http://stackoverflow.com/questions/5310255/how-can-i-force-pythonusing-win32com-to-create-a-new-instance-of-excel
                    results in error
                """
                clsid = pythoncom.CoCreateInstanceEx(CLSID, None, pythoncom.CLSCTX_SERVER,
                                            None, (pythoncom.IID_IDispatch,))[0]
                if gencache.is_readonly:
                    #fix for "freezed" app: py2exe.org/index.cgi/UsingEnsureDispatch
                    gencache.is_readonly = False
                    gencache.Rebuild()
                self.excel  = gencache.EnsureDispatch(clsid)

        else:
            self.excel  = Dispatch(
                            pythoncom.CoGetInterfaceAndReleaseStream( excel_id, pythoncom.IID_IDispatch )
                            )
    def __del__( self ):
        if self.excel is not None:
            print( 'Excel Quit in ExcelComClient.__del__' )
            # self.quit( alert=False )  # this does not work
            self.quit( alert=0 )
            print( 'Excel Quit ok' )

    def get_comid( self ):
        excel_id    = pythoncom.CoMarshalInterThreadInterfaceInStream( pythoncom.IID_IDispatch, self.excel )
        return excel_id

    def create_thread( self, target, args=[], kwargs={}, name='ExcelComClientThread' ):
        import threading
        thread      = threading.Thread(
                                    target=target,
                                    name=name,
                                    args=args,
                                    kwargs=kwargs
                                    )
        return thread

    def selection( self ):
        return self.excel.Selection

    def quit( self, alert=True ):
        self.excel.Application.DisplayAlerts = alert
        self.excel.Application.Quit()
        self.excel = None
        # print( 'Excel Quit ok' )

    def openWorkbook( self, path, save=True ):
        wb = Workbook_( self, path, save )
        return wb

    def openWorksheet( self, path ):
        ws = Worksheet( self, path )
        return ws

    def merge_books( self, input_books, merged_book, progress_cb=None, merged_cb=None ):
        if merged_book in input_books:
            raise Exception( "input_books must not include merged_book" )

        if os.path.exists( merged_book ):
            os.remove( merged_book )

        # Opening a new book causes a loss of book parameters
        # which are not copied by ws.Copy method.
        # So, it is better to filecopy to preserve ones in the
        # first book.
        copyfile( input_books[-1], merged_book )

        if progress_cb is not None:
            progress_cb(0)

        if len( input_books ) == 1:
            return

        out_wb= self.openWorkbook( merged_book ).workbook
        num_sheets = out_wb.Sheets.Count
        target_ws = out_wb.Sheets(num_sheets)
        # print( 'target_ws.Name=', target_ws.Name )

        for i, book in enumerate(input_books[:-1]):
            try:
                in_wb = self.openWorkbook( book ).workbook
            except:
                print( 'Skipped because of open error: ' + book )
                continue

            for ws in in_wb.Sheets:
                # print( "ws.Name", ws.Name )
                # ? After results in a new book
                ws.Copy( Before=target_ws )
            in_wb.Close( SaveChanges=False )
            if progress_cb is not None:
                progress_cb(i)

        target_ws.Delete()

        if merged_cb is not None:
            merged_cb( self, out_wb )

        out_wb.Sheets(1).Activate()
        out_wb.Save()
        out_wb.Close()

def merge_into_a_book( input_books, merged_book, excel_client=None, progress_cb=None ):
    CoInitialize()  # required in thread

    excel_client_ = ExcelComClient() if excel_client is None else excel_client

    """
    TODO: this doen't work!
    def freeze_pane_sheet1( excel_client, wb ):
        # workaourd to fix broken FreezePanes setting
        ws = wb.Worksheets(1)
        ws.Activate()
        excel_client.excel.Range( "D2" ).Select()
        aw = excel_client.excel.ActiveWindow
        aw.FreezePanes  = True
        aw.ScrollColumn = 16

    excel_client_.merge_books( input_books, merged_book, progress_cb=progress_cb, merged_cb=freeze_pane_sheet1 )
    """
    excel_client_.merge_books( input_books, merged_book, progress_cb=progress_cb )

    excel_client_ = None
    CoUninitialize()

"""
    Naming this class as Workbook_ is to distinguish Workbook  from openpyxl
"""
class Workbook_:
    def __init__( self, excel_client, path, save=True ):
        self.excel_client   = excel_client
        self.excel          = excel_client.excel
        if os.path.exists( path ):
            wb_ = self.excel.Workbooks.Open( path )
        else:
            wb_ = self.excel.Workbooks.Add()
            if save:
                wb_.SaveAs( path )

        self.workbook = wb_

class Worksheet:
    def __init__( self, excel_client, path ):
        self.excel_client   = excel_client
        self.excel          = excel_client.excel
        m = worksheet_path_re.match( path )
        if not m: raise Exception('Invalid path: ' + path )

        wb_path = m.group(1)
        ws_name = m.group(2)
        print( 'Worksheet:', wb_path, ws_name )

        self.workbook   = wb_ = self.excel_client.openWorkbook( wb_path ).workbook
        try:
            ws_ = wb_.Worksheets(ws_name)
        except:
            ws_ = wb_.Worksheets(1)
            ws_.Name = ws_name

        print( 'ws_.Name=', ws_.Name )
        self.worksheet  = ws_

    def draw_line( self, p1, p2 ):
        line = self.worksheet.Shapes.AddConnector( msoConnectorStraight, p1[0], p1[1], p2[0], p2[1] )
        line.Select()
        return line

    def draw_oval( self, p, a, b, color=None ):
        oval = self.worksheet.Shapes.AddShape( msoShapeOval, p[0], p[1], a, b )
        oval.Select()
        if color is not None:
            selection = self.excel.Selection
            selection.ShapeRange.Fill.ForeColor.RGB = color
        return oval

    def draw_right_brace( self, p, width, hight, rotation ):
        """
        Function AddShape(Type As MsoAutoShapeType, Left As Single, Top As Single, Width As Single, Height As Single) As Shape
        """
        brace = self.worksheet.Shapes.AddShape( msoShapeRightBrace, p[0], p[1], width, hight )
        brace.Select()
        self.excel.Selection.ShapeRange.Rotation = rotation
        return brace

    def draw_textbox( self, text, p, width, hight, visible=True ):
        textbox = self.worksheet.Shapes.AddTextbox( msoTextOrientationHorizontal, p[0], p[1], width, hight )
        textbox.Select()
        selection = self.excel.Selection
        selection.ShapeRange.Line.Visible = visible
        selection.ShapeRange(1).TextFrame2.TextRange.Characters.Text = text
        return textbox

    def freeze_panes( self, range_, scroll_column=None ):
        self.excel.Range( range_ ).Select()
        self.excel.ActiveWindow.FreezePanes = True
        if scroll_column is not None:
            self.excel.ActiveWindow.ScrollColumn = scroll_column

    def zoom( self, percentage ):
        self.excel.ActiveWindow.Zoom = percentage

    def get_num_charts( self ):
        return self.worksheet.ChartObjects().Count

    def get_chart( self, i ):
        if DEBUG:
            for chart in self.worksheet.ChartObjects():
                print( chart.Name )

        chartobj = self.worksheet.ChartObjects()[i]

        if DEBUG:
            print( 'chartobj.Chart=', chartobj.Chart )
            for ser in chartobj.Chart.SeriesCollection():
                print( ser.Name )

        return chartobj.Chart

    def get_chart_data_point( self, chart, i, j ):
        # <class 'numpy.int64'> can cause 'MemoryError: CreatingSafeArray'
        i_, j_ = int(i), int(j)
        point  = chart.SeriesCollection()[i_].Points()[j_]

        if DEBUG:
            print( 'point.Left=', point.Left )
            print( 'point.Top=', point.Top )

        return point

def compute_axis_max_value( digits, max_value ):
    n = int( np.ceil( np.log10( max_value ) ) )
    print( 'n=', n )
    scale_ = 10**( digits - 1 )
    v = np.ceil( max_value * scale_ ) / scale_
    print( 'digits=', digits, 'scale_=', scale_, 'v=', v )
    return v
